
package org.python.modules.csv;

import org.python.core.Py;
import org.python.core.PyException;
import org.python.core.PyIterator;
import org.python.core.PyList;
import org.python.core.PyNone;
import org.python.core.PyObject;
import org.python.core.PyString;
import org.python.core.PyType;
import org.python.expose.ExposedType;
import org.python.expose.ExposedGet;


// struct ReaderObj (_csv.c)
@ExposedType(name = "Reader")
public class PyReader extends PyIterator {
    public String __doc__ = "CSV reader\n" +
        "\n" +
        "Reader objects are responsible for reading and parsing tabular data\n" +
        "in CSV format.\n";

    // Parsing dialect
    @ExposedGet
    public PyDialect dialect;

    // Iterate over this for input lines
    private PyObject input_iter;

    // Field list for current record
    private PyList fields;

    // Current CSV parse state
    private int state;

    // Build current field in here
    private StringBuffer field;

    // did we have a parse error?
    private boolean had_parse_error;

    // csv_reader (_csv.c)
    public PyReader(PyObject[] args, String[] kwargs) {
        super();

        this.dialect = null;
        this.input_iter = null;
        this.fields = null;
        this.field = null;
        this.had_parse_error = false;
        this.state = _csv.START_RECORD;

        if (args.length < 1) {
            throw Py.TypeError("expected at least 1 argument, got 0");
        }

        // The "csvfile" parameter is the only positional argument that
        // must exist and is referred to as the input_iter in the source.
        // Py.iter() (rather unobviously!) throws a TypeError if input_iter
        // is not an iterator.
        input_iter = args[0];
        input_iter = Py.iter(input_iter, "argument 1 must be an iterator");

        // Dialect is responsible for extracting the dialect argument
        // and all the keyword parameters that make up "fmtparam".

        dialect = new PyDialect(args, kwargs);

        fields = new PyList();
        field = new StringBuffer();
    }


    // Reader_iternext (_csv.c)
    public PyObject __iternext__() {
        PyObject lineobj;
        PyObject fields;
        String line;

        do {
            lineobj = input_iter.__iternext__();
            if (lineobj == null) {
                // End of input OR exception
                if (this.field.length() != 0) {
                    throw new PyException(_csv.Error, "newline inside string");
                } else {
                    return null;
                }
            }

            if (this.had_parse_error) {
                this.fields = new PyList();
                this.state = _csv.START_RECORD;
                this.had_parse_error = false;
            }

            line = lineobj.toString();
            if (line == null) {
                return null;
            }
            if (line.contains("\0")) {
                this.had_parse_error = true;
                throw new PyException(_csv.Error, "string with NUL bytes");
            }
            // parsing below is based on _csv.c which
            // expects all strings to be terminated with
            // a nul byte. 
            line += "\0";

            // Process line of text - send '\n' to processing code to
            // represent end of line. End of line which is not at end of
            // string is an error.
            int idx = 0;
            while (idx < line.length() - 1) { // -1 to accomodate for nul
                char c = line.charAt(idx++);

                if (c == '\r') {
                    c = line.charAt(idx++);
                    if (c == '\0') {
                        // macintosh end of line
                        break;
                    }
                    if (c == '\n') {
                        c = line.charAt(idx++);
                        if (c == '\0') {
                            // DOS end of line
                            break;
                        }
                    }
                    this.had_parse_error = true;
                    throw new PyException(_csv.Error, "newline inside string");
                }
                if (c == '\n') {
                    c = line.charAt(idx++);
                    if (c == '\0') {
                        // unix end of line
                        break;
                    }
                    this.had_parse_error = true;
                    throw new PyException(_csv.Error, "newline inside string");
                }
                parse_process_char(c);
            }
            parse_process_char('\n');

        } while (this.state != _csv.START_RECORD);

        fields = this.fields;
        this.fields = new PyList();

        return fields;
    }

    private void parse_process_char(char c) {

        PyDialect dialect = this.dialect;

        switch (this.state) {

            case _csv.START_RECORD:
                // start of record
                if (c == '\n') {
                    // empty line - return []
                    break;
                }
                // normal character - handle as START_FIELD
                this.state = _csv.START_FIELD;
                // *** fallthru ***
            case _csv.START_FIELD:
                // expecting field
                if (c == '\n') {
                    // save empty field - return [fields]
                    parse_save_field();
                    this.state = _csv.START_RECORD;
                } else if (c == dialect.quotechar) {
                    // start quoted field
                    this.state = _csv.IN_QUOTED_FIELD;
                } else if (c == dialect.escapechar) {
                    // possible escaped character
                    this.state = _csv.ESCAPED_CHAR;
                } else if (c == ' ' && dialect.skipinitialspace) {
                    // ignore space at start of field
                    ;
                } else if (c == dialect.delimiter) {
                    // save empty field
                    parse_save_field();
                } else {
                    // begin new unquoted field
                    parse_add_char(c);
                    this.state = _csv.IN_FIELD;
                }
                break;

            case _csv.ESCAPED_CHAR:
                if (c != dialect.escapechar && c != dialect.delimiter && c != dialect.quotechar) {
                    parse_add_char(dialect.escapechar);
                }
                parse_add_char(c);
                this.state = _csv.IN_FIELD;
                break;

            case _csv.IN_FIELD:
                // in unquoted field
                if (c == '\n') {
                    // end of line - return [fields]
                    parse_save_field();
                    this.state = _csv.START_RECORD;
                } else if (c == dialect.escapechar) {
                    // possible escaped character
                    this.state = _csv.ESCAPED_CHAR;
                } else if (c == dialect.delimiter) {
                    // save field - wait for new field
                    parse_save_field();
                    this.state = _csv.START_FIELD;
                } else {
                    // normal character - save in field
                    parse_add_char(c);
                }
                break;

            case _csv.IN_QUOTED_FIELD:
                // in quoted field
                if (c == '\n') {
                    // end of line - save '\n' in field
                    parse_add_char('\n');
                } else if (c == dialect.escapechar) {
                    // Possible escape character
                    this.state = _csv.ESCAPE_IN_QUOTED_FIELD;
                } else if (c == dialect.quotechar) {
                    if (dialect.doublequote) {
                        // doublequote; " represented by ""
                        this.state = _csv.QUOTE_IN_QUOTED_FIELD;
                    } else {
                        // end of quote part of field
                        this.state = _csv.IN_FIELD;
                    }
                } else {
                    // normal character - save in field
                    parse_add_char(c);
                }
                break;

            case _csv.ESCAPE_IN_QUOTED_FIELD:
                if (c != dialect.escapechar && 
					c != dialect.delimiter && 
					c != dialect.quotechar) {
                    parse_add_char(dialect.escapechar);
                }
                parse_add_char(c);
                this.state = _csv.IN_QUOTED_FIELD;
                break;

            case _csv.QUOTE_IN_QUOTED_FIELD:
                // doublequote - seen a quote in an quoted field
                if (dialect.quoting != _csv.QUOTE_NONE && 
					c == dialect.quotechar) {
                    // save "" as "
                    parse_add_char(c);
                    this.state = _csv.IN_QUOTED_FIELD;
                } else if (c == dialect.delimiter) {
                    // save field - wait for new field
                    parse_save_field();
                    this.state = _csv.START_FIELD;
                } else if (c == '\n') {
                    // end of line - return [fields]
                    parse_save_field();
                    this.state = _csv.START_RECORD;
                } else if (!dialect.strict) {
                    parse_add_char(c);
                    this.state = _csv.IN_FIELD;
                } else {
                    // illegal
                    this.had_parse_error = true;
                    throw new PyException(_csv.Error,
                        "" + dialect.delimiter +
                        " expected after " +
                        dialect.quotechar);
                }
                break;
        }
    }

    private void parse_save_field() {
        PyObject field;

        field = new PyString(this.field.toString());
        if (field != null) {
            this.fields.append(field);
        }

        // fastest way to clear StringBuffer?
        this.field = new StringBuffer();
    }

    private void parse_add_char(char c) {
        this.field.append(c);
    }
}
