
package org.python.modules.csv;

import org.python.core.Py;
import org.python.core.PyException;
import org.python.core.PyObject;
import org.python.core.PyString;
import org.python.expose.ExposedType;
import org.python.expose.ExposedMethod;
import org.python.expose.ExposedGet;



// struct WriterObj (_csv.c)
@ExposedType(name = "Writer")
public class PyWriter extends PyObject {
    // write output lines to this file
    private PyObject writeline;

    // parsing dialect
    @ExposedGet
    public PyDialect dialect;

    // buffer for parser.join 
    private StringBuffer rec;

    // length of record
    private int rec_len;

    // number of fields in record
    private int num_fields;

    // whether field should be quoted during a join.  This is
    // instead of passing around the quoted parameter in all the join
    // methods in _csv.c
    private boolean quoted = false;

    public PyWriter(PyObject[] args, String[] keywords) {
        PyObject output_file;

        this.dialect = null;
        this.writeline = null;
        this.rec_len = 0;
        this.num_fields = 0;

        if (args.length < 1) {
            throw Py.TypeError("expected at least 1 argument, got 0");
        }
        output_file = args[0];
        try {
            this.writeline = output_file.__getattr__("write");
            if (this.writeline == null || !(this.writeline.isCallable())) {
                throw Py.TypeError("argument 1 must be an instance with a write method");
            }
        } catch(PyException ex) {
            if(Py.matchException(ex, Py.AttributeError)) {
                throw Py.TypeError("argument 1 must be an instance with a write method");
            }
        }

        // Dialect is responsible for extracting the dialect argument
        // and all the keyword parameters that make up "fmtparam".
        this.dialect = new PyDialect(args, keywords);
    }

    public static PyString __doc__writerows = new PyString(
            "writerows(sequence of sequences)\n" +
            "\n" +
            "Construct and write a series of sequences to a csv file.  Non-string\n" +
            "elements will be converted to string.");

    //csv_writerows (_csv.c)
    @ExposedMethod
    public void writerows(PyObject seqseq) {
        PyObject row_iter;
        PyObject row_obj;
        boolean result;

        row_iter = seqseq.__iter__();
        if (row_iter == null) {
            throw new PyException(_csv.Error,
                "writerows() argument must be iterable");
        }

        while ((row_obj = row_iter.__iternext__()) != null) {
            result = writerow(row_obj);
            if (!result) {
                break;
            }
        }
    }

    public static PyString __doc__writerow = new PyString(
            "writerow(sequence)\n" +
            "\n" +
            "Construct and write a CSV record from a sequence of fields.  Non-string\n" +
            "elements will be converted to string."
            );

    // csv_writerow (_csv.c)
    @ExposedMethod
    public boolean writerow(PyObject seq) {
        PyDialect dialect = this.dialect;
        int len, i;

        if (!seq.isSequenceType()) {
            throw new PyException(_csv.Error, "sequence expected");
        }

        len = seq.__len__();
        if (len < 0) {
            return false;
        }

        // Join all fields in internal buffer.
        join_reset();
        for (i = 0; i < len; i++) {
            PyObject field;
            boolean append_ok;

            quoted = false;

            field = seq.__getitem__(i);
            if (field == null) {
                return false;
            }

            if (dialect.quoting == _csv.QUOTE_NONNUMERIC) {

                try {
                    field.__float__();
                } catch(PyException ex) {
                    quoted = true;
                }
            }

            if (field instanceof PyString) {
                append_ok = join_append(field.toString(), len == 1);
            } else if (field == Py.None) {
                append_ok = join_append("", len == 1);
            } else {
                PyObject str;

                str = field.__str__();
                if (str == null) {
                    return false;
                }

                append_ok = join_append(str.toString(), len == 1);
            }
            if (!append_ok) {
                return false;
            }
        }

        // Add line terminator.
        if (!join_append_lineterminator()) {
            return false;
        }

        this.writeline.__call__(new PyString(rec.toString()));
        return true;
    }

    private void join_reset() {
        this.rec_len = 0;
        this.num_fields = 0;
        this.quoted = false;
        this.rec = new StringBuffer();
    }

    private boolean join_append_lineterminator() {
        this.rec.append(this.dialect.lineterminator.toString());
        return true;
    }

    private boolean join_append(String field, boolean quote_empty) {
        int rec_len;

        rec_len = join_append_data(field, quote_empty, false);
        if (rec_len < 0) {
            return false;
        }

        this.rec_len = join_append_data(field, quote_empty, true);
        this.num_fields++;

        return true;
    }

    /**
     * This method behaves differently depending on the value of copy_phase:
     * if copy_phase is false, then the method determines the new record
     * length. If copy_phase is true then the new field is appended to the
     * record.
     */
    private int join_append_data(String field, boolean quote_empty, boolean copy_phase) {
        PyDialect dialect = this.dialect;
        int i, rec_len;

        rec_len = this.rec_len;

        // If this is not the first field we need a field separator.
        if (this.num_fields > 0) {
            if (copy_phase) {
                this.rec.append(dialect.delimiter);
            }
            rec_len++;
        }

        // Handle preceding quote.
        switch (dialect.quoting) {
            case _csv.QUOTE_ALL:
                quoted = true;
                if (copy_phase) {
                    this.rec.append(dialect.quotechar);
                }
                rec_len++;
                break;
            case _csv.QUOTE_MINIMAL:
            case _csv.QUOTE_NONNUMERIC:
                // We only know about quoted in the copy phase.
                if (copy_phase && quoted) {
                    this.rec.append(dialect.quotechar);
                    rec_len++;
                }
                break;
            case _csv.QUOTE_NONE:
                break;
        }

        // parsing below is based on _csv.c which
        // expects all strings to be terminated with
        // a nul byte. 
        field += '\0';

        // Copy/count field data.
        for (i = 0;; i++) {
            char c = field.charAt(i);

            if (c == '\0') {
                break;
            }
            // If in doublequote mode we escape quote chars with a quote.
            if (dialect.quoting != _csv.QUOTE_NONE && c == dialect.quotechar
                    && dialect.doublequote) {
                if (copy_phase) {
                    this.rec.append(dialect.quotechar);
                }
                quoted = true;
                rec_len++;
                    }

            /*
             * Some special characters need to be escaped. If we have a
             * quote character switch to quoted field instead of escaping
             * individual characters.
             */
            if (!quoted && (c == dialect.delimiter || 
                        c == dialect.escapechar || c == '\n' || c == '\r')) {
                if (dialect.quoting != _csv.QUOTE_NONE) {
                    quoted = true;
                } else if (dialect.escapechar != '\0') { 
                    // Check _csv.c
                    if (copy_phase) {
                        this.rec.append(dialect.escapechar);
                    }
                    rec_len++;
                } else {
                    throw new PyException(_csv.Error, "delimiter must be quoted or escaped");
                }
                        }
            // Copy field character into record buffer.
            if (copy_phase) {
                this.rec.append(c);
            }
            rec_len++;
        }

        // If field is empty check if it needs to be quoted.
        if (i == 0 && quote_empty) {
            if (dialect.quoting == _csv.QUOTE_NONE) {
                throw new PyException(_csv.Error, 
                    "single empty field record must be quoted");

            } else {
                quoted = true;
            }
        }

        // Handle final quote character on field.
        if (quoted) {
            if (copy_phase) {
                this.rec.append(dialect.quotechar);
            } else {
                // Didn't know about leading quote until we found it
                // necessary in field data - compensate for it now.
                rec_len++;
            }
            rec_len++;
        }

        return rec_len;
    }

}

